@@ -15,14 +15,14 @@
 """The core data store and collection logic for beets.
 """
 import os
-import re
 import sys
 import logging
 import shlex
 import unicodedata
 import time
+import re
 from unidecode import unidecode
-from beets.mediafile import MediaFile, MutagenError
+from beets.mediafile import MediaFile, MutagenError, UnreadableFileError
 from beets import plugins
 from beets import util
 from beets.util import bytestring_path, syspath, normpath, samefile
@@ -39,6 +39,10 @@ log = logging.getLogger('beets')
 
 class PathQuery(dbcore.FieldQuery):
     """A query that matches all items under a given path."""
+
+    escape_re = re.compile(r'[\\_%]')
+    escape_char = '\\'
+
     def __init__(self, field, pattern, fast=True):
         super(PathQuery, self).__init__(field, pattern, fast)
 
@@ -52,19 +56,20 @@ class PathQuery(dbcore.FieldQuery):
             item.path.startswith(self.dir_path)
 
     def clause(self):
-        dir_pat = buffer(self.dir_path + '%')
+        escape = lambda m: self.escape_char + m.group(0)
+        dir_pattern = self.escape_re.sub(escape, self.dir_path)
+        dir_pattern = buffer(dir_pattern + '%')
         file_blob = buffer(self.file_path)
-        return '({0} = ?) || ({0} LIKE ?)'.format(self.field), \
-               (file_blob, dir_pat)
+        return '({0} = ?) || ({0} LIKE ? ESCAPE ?)'.format(self.field), \
+               (file_blob, dir_pattern, self.escape_char)
 
 
 # Library-specific field types.
 
-
-class DateType(types.Type):
-    sql = u'REAL'
+class DateType(types.Float):
+    # TODO representation should be `datetime` object
+    # TODO distinguish beetween date and time types
     query = dbcore.query.DateQuery
-    null = 0.0
 
     def format(self, value):
         return time.strftime(beets.config['time_format'].get(unicode),
@@ -87,6 +92,7 @@ class DateType(types.Type):
 class PathType(types.Type):
     sql = u'BLOB'
     query = PathQuery
+    model_type = bytes
 
     def format(self, value):
         return util.displayable_path(value)
@@ -107,21 +113,74 @@ class PathType(types.Type):
         else:
             return value
 
+    def from_sql(self, sql_value):
+        return self.normalize(sql_value)
+
+    def to_sql(self, value):
+        if isinstance(value, str):
+            value = buffer(value)
+        return value
+
+
+class MusicalKey(types.String):
+    """String representing the musical key of a song.
+
+    The standard format is C, Cm, C#, C#m, etc.
+    """
+    ENHARMONIC = {
+        r'db': 'c#',
+        r'eb': 'd#',
+        r'gb': 'f#',
+        r'ab': 'g#',
+        r'bb': 'a#',
+    }
+
+    def parse(self, key):
+        key = key.lower()
+        for flat, sharp in self.ENHARMONIC.items():
+            key = re.sub(flat, sharp, key)
+        key = re.sub(r'[\W\s]+minor', 'm', key)
+        return key.capitalize()
+
+    def normalize(self, key):
+        if key is None:
+            return None
+        else:
+            return self.parse(key)
+
+
+# Library-specific sort types.
+
+class SmartArtistSort(dbcore.query.Sort):
+    """Sort by artist (either album artist or track artist),
+    prioritizing the sort field over the raw field.
+    """
+    def __init__(self, model_cls, ascending=True):
+        self.album = model_cls is Album
+        self.ascending = ascending
+
+    def order_clause(self):
+        order = "ASC" if self.ascending else "DESC"
+        if self.album:
+            field = 'albumartist'
+        else:
+            field = 'artist'
+        return ('(CASE {0}_sort WHEN NULL THEN {0} '
+                'WHEN "" THEN {0} '
+                'ELSE {0}_sort END) {1}').format(field, order)
+
+    def sort(self, objs):
+        if self.album:
+            key = lambda a: a.albumartist_sort or a.albumartist
+        else:
+            key = lambda i: i.artist_sort or i.artist
+        return sorted(objs, key=key, reverse=not self.ascending)
+
 
 # Special path format key.
 PF_KEY_DEFAULT = 'default'
 
 
-# A little SQL utility.
-def _orelse(exp1, exp2):
-    """Generates an SQLite expression that evaluates to exp1 if exp1 is
-    non-null and non-empty or exp2 otherwise.
-    """
-    return ("""(CASE {0} WHEN NULL THEN {1}
-                         WHEN "" THEN {1}
-                         ELSE {0} END)""").format(exp1, exp2)
-
-
 # Exceptions.
 
 class FileOperationError(Exception):
@@ -169,7 +228,6 @@ class WriteError(FileOperationError):
 class LibModel(dbcore.Model):
     """Shared concrete functionality for Items and Albums.
     """
-    _bytes_keys = ('path', 'artpath')
 
     def _template_funcs(self):
         funcs = DefaultTemplateFunctions(self, self._db).functions()
@@ -189,25 +247,76 @@ class LibModel(dbcore.Model):
         plugins.send('database_change', lib=self._db)
 
 
+class FormattedItemMapping(dbcore.db.FormattedMapping):
+    """Add lookup for album-level fields.
+
+    Album-level fields take precedence if `for_path` is true.
+    """
+
+    def __init__(self, item, for_path=False):
+        super(FormattedItemMapping, self).__init__(item, for_path)
+        self.album = item.get_album()
+        self.album_keys = []
+        if self.album:
+            for key in self.album.keys(True):
+                if key in Album.item_keys or key not in item._fields.keys():
+                    self.album_keys.append(key)
+        self.all_keys = set(self.model_keys).union(self.album_keys)
+
+    def _get(self, key):
+        """Get the value for a key, either from the album or the item.
+        Raise a KeyError for invalid keys.
+        """
+        if self.for_path and key in self.album_keys:
+            return self._get_formatted(self.album, key)
+        elif key in self.model_keys:
+            return self._get_formatted(self.model, key)
+        elif key in self.album_keys:
+            return self._get_formatted(self.album, key)
+        else:
+            raise KeyError(key)
+
+    def __getitem__(self, key):
+        """Get the value for a key. Certain unset values are remapped.
+        """
+        value = self._get(key)
+
+        # `artist` and `albumartist` fields fall back to one another.
+        # This is helpful in path formats when the album artist is unset
+        # on as-is imports.
+        if key == 'artist' and not value:
+            return self._get('albumartist')
+        elif key == 'albumartist' and not value:
+            return self._get('artist')
+        else:
+            return value
+
+    def __iter__(self):
+        return iter(self.all_keys)
+
+    def __len__(self):
+        return len(self.all_keys)
+
+
 class Item(LibModel):
     _table = 'items'
     _flex_table = 'item_attributes'
     _fields = {
-        'id':       types.Id(True),
+        'id':       types.PRIMARY_ID,
         'path':     PathType(),
-        'album_id': types.Id(False),
+        'album_id': types.FOREIGN_ID,
 
-        'title':                types.String(),
-        'artist':               types.String(),
-        'artist_sort':          types.String(),
-        'artist_credit':        types.String(),
-        'album':                types.String(),
-        'albumartist':          types.String(),
-        'albumartist_sort':     types.String(),
-        'albumartist_credit':   types.String(),
-        'genre':                types.String(),
-        'composer':             types.String(),
-        'grouping':             types.String(),
+        'title':                types.STRING,
+        'artist':               types.STRING,
+        'artist_sort':          types.STRING,
+        'artist_credit':        types.STRING,
+        'album':                types.STRING,
+        'albumartist':          types.STRING,
+        'albumartist_sort':     types.STRING,
+        'albumartist_credit':   types.STRING,
+        'genre':                types.STRING,
+        'composer':             types.STRING,
+        'grouping':             types.STRING,
         'year':                 types.PaddedInt(4),
         'month':                types.PaddedInt(2),
         'day':                  types.PaddedInt(2),
@@ -215,44 +324,44 @@ class Item(LibModel):
         'tracktotal':           types.PaddedInt(2),
         'disc':                 types.PaddedInt(2),
         'disctotal':            types.PaddedInt(2),
-        'lyrics':               types.String(),
-        'comments':             types.String(),
-        'bpm':                  types.Integer(),
-        'comp':                 types.Boolean(),
-        'mb_trackid':           types.String(),
-        'mb_albumid':           types.String(),
-        'mb_artistid':          types.String(),
-        'mb_albumartistid':     types.String(),
-        'albumtype':            types.String(),
-        'label':                types.String(),
-        'acoustid_fingerprint': types.String(),
-        'acoustid_id':          types.String(),
-        'mb_releasegroupid':    types.String(),
-        'asin':                 types.String(),
-        'catalognum':           types.String(),
-        'script':               types.String(),
-        'language':             types.String(),
-        'country':              types.String(),
-        'albumstatus':          types.String(),
-        'media':                types.String(),
-        'albumdisambig':        types.String(),
-        'disctitle':            types.String(),
-        'encoder':              types.String(),
-        'rg_track_gain':        types.NullFloat(),
-        'rg_track_peak':        types.NullFloat(),
-        'rg_album_gain':        types.NullFloat(),
-        'rg_album_peak':        types.NullFloat(),
+        'lyrics':               types.STRING,
+        'comments':             types.STRING,
+        'bpm':                  types.INTEGER,
+        'comp':                 types.BOOLEAN,
+        'mb_trackid':           types.STRING,
+        'mb_albumid':           types.STRING,
+        'mb_artistid':          types.STRING,
+        'mb_albumartistid':     types.STRING,
+        'albumtype':            types.STRING,
+        'label':                types.STRING,
+        'acoustid_fingerprint': types.STRING,
+        'acoustid_id':          types.STRING,
+        'mb_releasegroupid':    types.STRING,
+        'asin':                 types.STRING,
+        'catalognum':           types.STRING,
+        'script':               types.STRING,
+        'language':             types.STRING,
+        'country':              types.STRING,
+        'albumstatus':          types.STRING,
+        'media':                types.STRING,
+        'albumdisambig':        types.STRING,
+        'disctitle':            types.STRING,
+        'encoder':              types.STRING,
+        'rg_track_gain':        types.NULL_FLOAT,
+        'rg_track_peak':        types.NULL_FLOAT,
+        'rg_album_gain':        types.NULL_FLOAT,
+        'rg_album_peak':        types.NULL_FLOAT,
         'original_year':        types.PaddedInt(4),
         'original_month':       types.PaddedInt(2),
         'original_day':         types.PaddedInt(2),
-        'initial_key':          types.MusicalKey(),
+        'initial_key':          MusicalKey(),
 
-        'length':      types.Float(),
+        'length':      types.FLOAT,
         'bitrate':     types.ScaledInt(1000, u'kbps'),
-        'format':      types.String(),
+        'format':      types.STRING,
         'samplerate':  types.ScaledInt(1000, u'kHz'),
-        'bitdepth':    types.Integer(),
-        'channels':    types.Integer(),
+        'bitdepth':    types.INTEGER,
+        'channels':    types.INTEGER,
         'mtime':       DateType(),
         'added':       DateType(),
     }
@@ -269,6 +378,10 @@ class Item(LibModel):
     `write`.
     """
 
+    _formatter = FormattedItemMapping
+
+    _sorts = {'artist': SmartArtistSort}
+
     @classmethod
     def _getters(cls):
         getters = plugins.item_field_getters()
@@ -322,7 +435,7 @@ class Item(LibModel):
     def read(self, read_path=None):
         """Read the metadata from the associated file.
 
-        If ``read_path`` is specified, read metadata from that file
+        If `read_path` is specified, read metadata from that file
         instead. Updates all the properties in `_media_fields`
         from the media file.
 
@@ -334,7 +447,7 @@ class Item(LibModel):
             read_path = normpath(read_path)
         try:
             mediafile = MediaFile(syspath(read_path))
-        except (OSError, IOError) as exc:
+        except (OSError, IOError, UnreadableFileError) as exc:
             raise ReadError(read_path, exc)
 
         for key in self._media_fields:
@@ -365,16 +478,19 @@ class Item(LibModel):
             path = self.path
         else:
             path = normpath(path)
+
+        tags = dict(self)
+        plugins.send('write', item=self, path=path, tags=tags)
+
         try:
-            mediafile = MediaFile(path)
-        except (OSError, IOError) as exc:
+            mediafile = MediaFile(syspath(path),
+                                  id3v23=beets.config['id3v23'].get(bool))
+        except (OSError, IOError, UnreadableFileError) as exc:
             raise ReadError(self.path, exc)
 
-        plugins.send('write', item=self, path=path)
-
-        mediafile.update(self)
+        mediafile.update(tags)
         try:
-            mediafile.save(id3v23=beets.config['id3v23'].get(bool))
+            mediafile.save()
         except (OSError, IOError, MutagenError) as exc:
             raise WriteError(self.path, exc)
 
@@ -396,9 +512,25 @@ class Item(LibModel):
             log.error(exc)
             return False
 
+    def try_sync(self, write=None):
+        """Synchronize the item with the database and the media file
+        tags, updating them with this object's current state.
+
+        By default, the current `path` for the item is used to write
+        tags. If `write` is `False`, no tags are written. If `write` is
+        a path, tags are written to that file instead.
+
+        Similar to calling :meth:`write` and :meth:`store`.
+        """
+        if write is True:
+            write = None
+        if write is not False:
+            self.try_write(path=write)
+        self.store()
+
     # Files themselves.
 
-    def move_file(self, dest, copy=False):
+    def move_file(self, dest, copy=False, link=False):
         """Moves or copies the item's file, updating the path value if
         the move succeeds. If a file exists at ``dest``, then it is
         slightly modified to be unique.
@@ -409,6 +541,10 @@ class Item(LibModel):
             util.copy(self.path, dest)
             plugins.send("item_copied", item=self, source=self.path,
                          destination=dest)
+        elif link:
+            util.link(self.path, dest)
+            plugins.send("item_linked", item=self, source=self.path,
+                         destination=dest)
         else:
             plugins.send("before_item_moved", item=self, source=self.path,
                          destination=dest)
@@ -450,13 +586,14 @@ class Item(LibModel):
 
         self._db._memotable = {}
 
-    def move(self, copy=False, basedir=None, with_album=True):
+    def move(self, copy=False, link=False, basedir=None, with_album=True):
         """Move the item to its designated location within the library
         directory (provided by destination()). Subdirectories are
         created as needed. If the operation succeeds, the item's path
         field is updated to reflect the new location.
 
-        If copy is True, moving the file is copied rather than moved.
+        If `copy` is true, moving the file is copied rather than moved.
+        Similarly, `link` creates a symlink instead.
 
         basedir overrides the library base directory for the
         destination.
@@ -478,7 +615,7 @@ class Item(LibModel):
 
         # Perform the move and store the change.
         old_path = self.path
-        self.move_file(dest, copy)
+        self.move_file(dest, copy, link)
         self.store()
 
         # If this item is in an album, move its art.
@@ -494,12 +631,6 @@ class Item(LibModel):
 
     # Templating.
 
-    def _formatted_mapping(self, for_path=False):
-        """Get a mapping containing string-formatted values from either
-        this item or the associated album, if any.
-        """
-        return FormattedItemMapping(self, for_path)
-
     def destination(self, fragment=False, basedir=None, platform=None,
                     path_formats=None):
         """Returns the path in the library directory designated for the
@@ -519,7 +650,7 @@ class Item(LibModel):
         for query, path_format in path_formats:
             if query == PF_KEY_DEFAULT:
                 continue
-            query = get_query(query, type(self))
+            query, _ = parse_query_string(query, type(self))
             if query.match(self):
                 # The query matches the item! Use the corresponding path
                 # format.
@@ -544,8 +675,13 @@ class Item(LibModel):
             subpath = unicodedata.normalize('NFD', subpath)
         else:
             subpath = unicodedata.normalize('NFC', subpath)
+
+        if beets.config['asciify_paths']:
+            subpath = unidecode(subpath)
+
         # Truncate components and remove forbidden characters.
         subpath = util.sanitize_path(subpath, self._db.replacements)
+
         # Encode for the filesystem.
         if not fragment:
             subpath = bytestring_path(subpath)
@@ -570,48 +706,6 @@ class Item(LibModel):
             return normpath(os.path.join(basedir, subpath))
 
 
-class FormattedItemMapping(dbcore.db.FormattedMapping):
-    """A `dict`-like formatted view of an item that inherits album fields.
-
-    The accessor ``mapping[key]`` returns the formated version of either
-    ``item[key]`` or ``album[key]``. Here `album` is the album
-    associated to `item` if it exists.
-    """
-    # FIXME This class should be in the same module as `FormattedMapping`
-
-    def __init__(self, item, for_path=False):
-        self.for_path = for_path
-        self.model = item
-        self.model_keys = item.keys(True)
-        self.album = item.get_album()
-        self.album_keys = []
-        if self.album:
-            for key in self.album.keys(True):
-                if key in Album.item_keys or key not in item._fields.keys():
-                    self.album_keys.append(key)
-
-    def get(self, key):
-        if key in self.album_keys:
-            return self.album._get_formatted(key, self.for_path)
-        elif key in self.model_keys:
-            return self.model._get_formatted(key, self.for_path)
-        else:
-            raise KeyError(key)
-
-    def __getitem__(self, key):
-        value = self.get(key)
-
-        if key == 'artist' and not value:
-            return self.get('albumartist')
-        if key == 'albumartist' and not value:
-            return self.get('artist')
-
-        return value
-
-    def __contains__(self, key):
-        return key in self.model_keys or key in self.album_keys
-
-
 class Album(LibModel):
     """Provides access to information about albums stored in a
     library. Reflects the library's "albums" table, including album
@@ -619,37 +713,37 @@ class Album(LibModel):
     """
     _table = 'albums'
     _flex_table = 'album_attributes'
+    _always_dirty = True
     _fields = {
-        'id':      types.Id(True),
+        'id':      types.PRIMARY_ID,
         'artpath': PathType(),
         'added':   DateType(),
 
-        'albumartist':        types.String(),
-        'albumartist_sort':   types.String(),
-        'albumartist_credit': types.String(),
-        'album':              types.String(),
-        'genre':              types.String(),
+        'albumartist':        types.STRING,
+        'albumartist_sort':   types.STRING,
+        'albumartist_credit': types.STRING,
+        'album':              types.STRING,
+        'genre':              types.STRING,
         'year':               types.PaddedInt(4),
         'month':              types.PaddedInt(2),
         'day':                types.PaddedInt(2),
         'tracktotal':         types.PaddedInt(2),
         'disctotal':          types.PaddedInt(2),
-        'comp':               types.Boolean(),
-        'mb_albumid':         types.String(),
-        'mb_albumartistid':   types.String(),
-        'albumtype':          types.String(),
-        'label':              types.String(),
-        'mb_releasegroupid':  types.String(),
-        'asin':               types.String(),
-        'catalognum':         types.String(),
-        'script':             types.String(),
-        'language':           types.String(),
-        'country':            types.String(),
-        'albumstatus':        types.String(),
-        'media':              types.String(),
-        'albumdisambig':      types.String(),
-        'rg_album_gain':      types.NullFloat(),
-        'rg_album_peak':      types.NullFloat(),
+        'comp':               types.BOOLEAN,
+        'mb_albumid':         types.STRING,
+        'mb_albumartistid':   types.STRING,
+        'albumtype':          types.STRING,
+        'label':              types.STRING,
+        'mb_releasegroupid':  types.STRING,
+        'asin':               types.STRING,
+        'catalognum':         types.STRING,
+        'script':             types.STRING,
+        'language':           types.STRING,
+        'country':            types.STRING,
+        'albumstatus':        types.STRING,
+        'albumdisambig':      types.STRING,
+        'rg_album_gain':      types.NULL_FLOAT,
+        'rg_album_peak':      types.NULL_FLOAT,
         'original_year':      types.PaddedInt(4),
         'original_month':     types.PaddedInt(2),
         'original_day':       types.PaddedInt(2),
@@ -657,6 +751,11 @@ class Album(LibModel):
 
     _search_fields = ('album', 'albumartist', 'genre')
 
+    _sorts = {
+        'albumartist': SmartArtistSort,
+        'artist': SmartArtistSort,
+    }
+
     item_keys = [
         'added',
         'albumartist',
@@ -681,7 +780,6 @@ class Album(LibModel):
         'language',
         'country',
         'albumstatus',
-        'media',
         'albumdisambig',
         'rg_album_gain',
         'rg_album_peak',
@@ -726,7 +824,7 @@ class Album(LibModel):
             for item in self.items():
                 item.remove(delete, False)
 
-    def move_art(self, copy=False):
+    def move_art(self, copy=False, link=False):
         """Move or copy any existing album art so that it remains in the
         same directory as the items.
         """
@@ -739,9 +837,13 @@ class Album(LibModel):
             return
 
         new_art = util.unique_path(new_art)
-        log.debug('moving album art %s to %s' % (old_art, new_art))
+        log.debug(u'moving album art {0} to {1}'
+                  .format(util.displayable_path(old_art),
+                          util.displayable_path(new_art)))
         if copy:
             util.copy(old_art, new_art)
+        elif link:
+            util.link(old_art, new_art)
         else:
             util.move(old_art, new_art)
         self.artpath = new_art
@@ -751,7 +853,7 @@ class Album(LibModel):
             util.prune_dirs(os.path.dirname(old_art),
                             self._db.directory)
 
-    def move(self, copy=False, basedir=None):
+    def move(self, copy=False, link=False, basedir=None):
         """Moves (or copies) all items to their destination. Any album
         art moves along with them. basedir overrides the library base
         directory for the destination. The album is stored to the
@@ -766,10 +868,10 @@ class Album(LibModel):
         # Move items.
         items = list(self.items())
         for item in items:
-            item.move(copy, basedir=basedir, with_album=False)
+            item.move(copy, link, basedir=basedir, with_album=False)
 
         # Move art.
-        self.move_art(copy)
+        self.move_art(copy, link)
         self.store()
 
     def item_dir(self):
@@ -795,6 +897,8 @@ class Album(LibModel):
 
         filename_tmpl = Template(beets.config['art_filename'].get(unicode))
         subpath = self.evaluate_template(filename_tmpl, True)
+        if beets.config['asciify_paths']:
+            subpath = unidecode(subpath)
         subpath = util.sanitize_path(subpath,
                                      replacements=self._db.replacements)
         subpath = bytestring_path(subpath)
@@ -849,139 +953,68 @@ class Album(LibModel):
                         item[key] = value
                     item.store()
 
+    def try_sync(self, write=True):
+        """Synchronize the album and its items with the database and
+        their files by updating them with this object's current state.
 
-# Query construction and parsing helpers.
-
-PARSE_QUERY_PART_REGEX = re.compile(
-    # Non-capturing optional segment for the keyword.
-    r'(?:'
-    r'(\S+?)'    # The field key.
-    r'(?<!\\):'  # Unescaped :
-    r')?'
-
-    r'(.*)',         # The term itself.
-
-    re.I  # Case-insensitive.
-)
+        `write` indicates whether to write tags to the item files.
+        """
+        self.store()
+        for item in self.items():
+            item.try_sync(bool(write))
 
 
-def parse_query_part(part, query_classes={}, prefixes={},
-                     default_class=dbcore.query.SubstringQuery):
-    """Take a query in the form of a key/value pair separated by a
-    colon and return a tuple of `(key, value, cls)`. `key` may be None,
-    indicating that any field may be matched. `cls` is a subclass of
-    `FieldQuery`.
+# Query construction helpers.
 
-    The optional `query_classes` parameter maps field names to default
-    query types; `default_class` is the fallback. `prefixes` is a map
-    from query prefix markers and query types. Prefix-indicated queries
-    take precedence over type-based queries.
+def parse_query_parts(parts, model_cls):
+    """Given a beets query string as a list of components, return the
+    `Query` and `Sort` they represent.
 
-    To determine the query class, two factors are used: prefixes and
-    field types. For example, the colon prefix denotes a regular
-    expression query and a type map might provide a special kind of
-    query for numeric values. If neither a prefix nor a specific query
-    class is available, `default_class` is used.
-
-    For instance,
-    'stapler' -> (None, 'stapler', SubstringQuery)
-    'color:red' -> ('color', 'red', SubstringQuery)
-    ':^Quiet' -> (None, '^Quiet', RegexpQuery)
-    'color::b..e' -> ('color', 'b..e', RegexpQuery)
-
-    Prefixes may be "escaped" with a backslash to disable the keying
-    behavior.
+    Like `dbcore.parse_sorted_query`, with beets query prefixes and
+    special path query detection.
     """
-    part = part.strip()
-    match = PARSE_QUERY_PART_REGEX.match(part)
-
-    assert match  # Regex should always match.
-    key = match.group(1)
-    term = match.group(2).replace('\:', ':')
-
-    # Match the search term against the list of prefixes.
-    for pre, query_class in prefixes.items():
-        if term.startswith(pre):
-            return key, term[len(pre):], query_class
-
-    # No matching prefix: use type-based or fallback/default query.
-    query_class = query_classes.get(key, default_class)
-    return key, term, query_class
-
-
-def construct_query_part(query_part, model_cls):
-    """Create a query from a single query component, `query_part`, for
-    querying instances of `model_cls`. Return a `Query` instance.
-    """
-    # Shortcut for empty query parts.
-    if not query_part:
-        return dbcore.query.TrueQuery()
-
-    # Set up and parse the string.
-    query_classes = dict((k, t.query) for (k, t) in model_cls._fields.items())
+    # Get query types and their prefix characters.
     prefixes = {':': dbcore.query.RegexpQuery}
     prefixes.update(plugins.queries())
-    key, pattern, query_class = \
-        parse_query_part(query_part, query_classes, prefixes)
 
-    # No key specified.
-    if key is None:
-        if os.sep in pattern and 'path' in model_cls._fields:
-            # This looks like a path.
-            return PathQuery('path', pattern)
-        elif issubclass(query_class, dbcore.FieldQuery):
-            # The query type matches a specific field, but none was
-            # specified. So we use a version of the query that matches
-            # any field.
-            return dbcore.query.AnyFieldQuery(pattern,
-                                              model_cls._search_fields,
-                                              query_class)
-        else:
-            # Other query type.
-            return query_class(pattern)
-
-    key = key.lower()
-    return query_class(key.lower(), pattern, key in model_cls._fields)
-
-
-def query_from_strings(query_cls, model_cls, query_parts):
-    """Creates a collection query of type `query_cls` from a list of
-    strings in the format used by parse_query_part. `model_cls`
-    determines how queries are constructed from strings.
-    """
-    subqueries = []
-    for part in query_parts:
-        subqueries.append(construct_query_part(part, model_cls))
-    if not subqueries:  # No terms in query.
-        subqueries = [dbcore.query.TrueQuery()]
-    return query_cls(subqueries)
-
-
-def get_query(val, model_cls):
-    """Takes a value which may be None, a query string, a query string
-    list, or a Query object, and returns a suitable Query object.
-    `model_cls` is the subclass of Model indicating which entity this
-    is a query for (i.e., Album or Item) and is used to determine which
-    fields are searched.
-    """
-    # Convert a single string into a list of space-separated
-    # criteria.
-    if isinstance(val, basestring):
-        # A bug in Python < 2.7.3 prevents correct shlex splitting of
-        # Unicode strings.
-        # http://bugs.python.org/issue6988
-        if isinstance(val, unicode):
-            val = val.encode('utf8')
-        val = [s.decode('utf8') for s in shlex.split(val)]
-
-    if val is None:
-        return dbcore.query.TrueQuery()
-    elif isinstance(val, list) or isinstance(val, tuple):
-        return query_from_strings(dbcore.AndQuery, model_cls, val)
-    elif isinstance(val, dbcore.Query):
-        return val
+    # Special-case path-like queries, which are non-field queries
+    # containing path separators (/).
+    if 'path' in model_cls._fields:
+        path_parts = []
+        non_path_parts = []
+        for s in parts:
+            if s.find(os.sep, 0, s.find(':')) != -1:
+                # Separator precedes colon.
+                path_parts.append(s)
+            else:
+                non_path_parts.append(s)
     else:
-        raise ValueError('query must be None or have type Query or str')
+        path_parts = ()
+        non_path_parts = parts
+
+    query, sort = dbcore.parse_sorted_query(
+        model_cls, non_path_parts, prefixes
+    )
+
+    # Add path queries to aggregate query.
+    if path_parts:
+        query.subqueries += [PathQuery('path', s) for s in path_parts]
+    return query, sort
+
+
+def parse_query_string(s, model_cls):
+    """Given a beets query string, return the `Query` and `Sort` they
+    represent.
+
+    The string is split into components using shell-like syntax.
+    """
+    # A bug in Python < 2.7.3 prevents correct shlex splitting of
+    # Unicode strings.
+    # http://bugs.python.org/issue6988
+    if isinstance(s, unicode):
+        s = s.encode('utf8')
+    parts = [p.decode('utf8') for p in shlex.split(s)]
+    return parse_query_parts(parts, model_cls)
 
 
 # The Library: interface to the database.
@@ -1045,30 +1078,41 @@ class Library(dbcore.Database):
 
     # Querying.
 
-    def _fetch(self, model_cls, query, order_by=None):
-        """Parse a query and fetch.
+    def _fetch(self, model_cls, query, sort=None):
+        """Parse a query and fetch. If a order specification is present
+        in the query string the `sort` argument is ignored.
         """
+        # Parse the query, if necessary.
+        parsed_sort = None
+        if isinstance(query, basestring):
+            query, parsed_sort = parse_query_string(query, model_cls)
+        elif isinstance(query, (list, tuple)):
+            query, parsed_sort = parse_query_parts(query, model_cls)
+
+        # Any non-null sort specified by the parsed query overrides the
+        # provided sort.
+        if parsed_sort and not isinstance(parsed_sort, dbcore.query.NullSort):
+            sort = parsed_sort
+
         return super(Library, self)._fetch(
-            model_cls, get_query(query, model_cls), order_by
+            model_cls, query, sort
         )
 
-    def albums(self, query=None):
-        """Get a sorted list of :class:`Album` objects matching the
-        given query.
+    def albums(self, query=None, sort=None):
+        """Get :class:`Album` objects matching the query.
         """
-        order = '{0}, album'.format(
-            _orelse("albumartist_sort", "albumartist")
+        sort = sort or dbcore.sort_from_strings(
+            Album, beets.config['sort_album'].as_str_seq()
         )
-        return self._fetch(Album, query, order)
+        return self._fetch(Album, query, sort)
 
-    def items(self, query=None):
-        """Get a sorted list of :class:`Item` objects matching the given
-        query.
+    def items(self, query=None, sort=None):
+        """Get :class:`Item` objects matching the query.
         """
-        order = '{0}, album, disc, track'.format(
-            _orelse("artist_sort", "artist")
+        sort = sort or dbcore.sort_from_strings(
+            Item, beets.config['sort_item'].as_str_seq()
         )
-        return self._fetch(Item, query, order)
+        return self._fetch(Item, query, sort)
 
     # Convenience accessors.
 
@@ -1243,7 +1287,7 @@ class DefaultTemplateFunctions(object):
             return res
 
         # Flatten disambiguation value into a string.
-        disam_value = album._get_formatted(disambiguator, True)
+        disam_value = album.formatted(True).get(disambiguator)
         res = u' [{0}]'.format(disam_value)
         self.lib._memotable[memokey] = res
         return res
