@@ -40,6 +40,7 @@ import mutagen.mp4
 import mutagen.flac
 import mutagen.monkeysaudio
 import mutagen.asf
+import mutagen.aiff
 import datetime
 import re
 import base64
@@ -51,6 +52,9 @@ import logging
 import traceback
 import enum
 
+from beets.util import displayable_path
+
+
 __all__ = ['UnreadableFileError', 'FileTypeError', 'MediaFile']
 
 log = logging.getLogger('beets')
@@ -67,27 +71,40 @@ TYPES = {
     'wv':   'WavPack',
     'mpc':  'Musepack',
     'asf':  'Windows Media',
+    'aiff': 'AIFF',
 }
 
 
 # Exceptions.
 
 class UnreadableFileError(Exception):
-    """Indicates a file that MediaFile can't read.
+    """Mutagen is not able to extract information from the file.
     """
-    pass
+    def __init__(self, path):
+        Exception.__init__(self, displayable_path(path))
 
 
 class FileTypeError(UnreadableFileError):
-    """Raised for files that don't seem to have a type MediaFile
-    supports.
+    """Reading this type of file is not supported.
+
+    If passed the `mutagen_type` argument this indicates that the
+    mutagen type is not supported by `Mediafile`.
     """
-    pass
+    def __init__(self, path, mutagen_type=None):
+        path = displayable_path(path)
+        if mutagen_type is None:
+            msg = path
+        else:
+            msg = u'{0}: of mutagen type {1}'.format(path, mutagen_type)
+        Exception.__init__(self, msg)
 
 
 class MutagenError(UnreadableFileError):
     """Raised when Mutagen fails unexpectedly---probably due to a bug.
     """
+    def __init__(self, path, mutagen_exc):
+        msg = u'{0}: {1}'.format(displayable_path(path), mutagen_exc)
+        Exception.__init__(self, msg)
 
 
 # Utility.
@@ -137,11 +154,12 @@ def _safe_cast(out_type, val):
         else:
             if not isinstance(val, basestring):
                 val = unicode(val)
-            val = re.match(r'[\+-]?[0-9\.]*', val.strip()).group(0)
-            if not val:
-                return 0.0
-            else:
-                return float(val)
+            match = re.match(r'[\+-]?[0-9\.]+', val.strip())
+            if match:
+                val = match.group(0)
+                if val:
+                    return float(val)
+            return 0.0
 
     else:
         return val
@@ -393,7 +411,7 @@ class StorageStyle(object):
         """
         try:
             return mutagen_file[self.key][0]
-        except KeyError:
+        except (KeyError, IndexError):
             return None
 
     def deserialize(self, mutagen_value):
@@ -635,7 +653,7 @@ class MP4ImageStorageStyle(MP4ListStorageStyle):
 class MP3StorageStyle(StorageStyle):
     """Store data in ID3 frames.
     """
-    formats = ['MP3']
+    formats = ['MP3', 'AIFF']
 
     def __init__(self, key, id3_lang=None, **kwargs):
         """Create a new ID3 storage style. `id3_lang` is the value for
@@ -647,7 +665,7 @@ class MP3StorageStyle(StorageStyle):
     def fetch(self, mutagen_file):
         try:
             return mutagen_file[self.key].text[0]
-        except KeyError:
+        except (KeyError, IndexError):
             return None
 
     def store(self, mutagen_file, value):
@@ -803,6 +821,9 @@ class MP3ImageStorageStyle(ListStorageStyle, MP3StorageStyle):
     def store(self, mutagen_file, frames):
         mutagen_file.tags.setall(self.key, frames)
 
+    def delete(self, mutagen_file):
+        mutagen_file.tags.delall(self.key)
+
     def serialize(self, image):
         """Return an APIC frame populated with data from ``image``.
         """
@@ -850,7 +871,7 @@ class VorbisImageStorageStyle(ListStorageStyle):
     base64-encoded. Values are `Image` objects.
     """
     formats = ['OggOpus', 'OggTheora', 'OggSpeex', 'OggVorbis',
-               'OggFlac', 'APEv2File', 'WavPack', 'Musepack', 'MonkeysAudio']
+               'OggFlac']
 
     def __init__(self):
         super(VorbisImageStorageStyle, self).__init__(
@@ -926,6 +947,79 @@ class FlacImageStorageStyle(ListStorageStyle):
         pic.desc = image.desc or u''
         return pic
 
+    def delete(self, mutagen_file):
+        """Remove all images from the file.
+        """
+        mutagen_file.clear_pictures()
+
+
+class APEv2ImageStorageStyle(ListStorageStyle):
+    """Store images in APEv2 tags. Values are `Image` objects.
+    """
+    formats = ['APEv2File', 'WavPack', 'Musepack', 'MonkeysAudio', 'OptimFROG']
+
+    TAG_NAMES = {
+        ImageType.other: 'Cover Art (other)',
+        ImageType.icon: 'Cover Art (icon)',
+        ImageType.other_icon: 'Cover Art (other icon)',
+        ImageType.front: 'Cover Art (front)',
+        ImageType.back: 'Cover Art (back)',
+        ImageType.leaflet: 'Cover Art (leaflet)',
+        ImageType.media: 'Cover Art (media)',
+        ImageType.lead_artist: 'Cover Art (lead)',
+        ImageType.artist: 'Cover Art (artist)',
+        ImageType.conductor: 'Cover Art (conductor)',
+        ImageType.group: 'Cover Art (band)',
+        ImageType.composer: 'Cover Art (composer)',
+        ImageType.lyricist: 'Cover Art (lyricist)',
+        ImageType.recording_location: 'Cover Art (studio)',
+        ImageType.recording_session: 'Cover Art (recording)',
+        ImageType.performance: 'Cover Art (performance)',
+        ImageType.screen_capture: 'Cover Art (movie scene)',
+        ImageType.fish: 'Cover Art (colored fish)',
+        ImageType.illustration: 'Cover Art (illustration)',
+        ImageType.artist_logo: 'Cover Art (band logo)',
+        ImageType.publisher_logo: 'Cover Art (publisher logo)',
+    }
+
+    def __init__(self):
+        super(APEv2ImageStorageStyle, self).__init__(key='')
+
+    def fetch(self, mutagen_file):
+        images = []
+        for cover_type, cover_tag in self.TAG_NAMES.items():
+            try:
+                frame = mutagen_file[cover_tag]
+                text_delimiter_index = frame.value.find('\x00')
+                comment = frame.value[0:text_delimiter_index] \
+                    if text_delimiter_index > 0 else None
+                image_data = frame.value[text_delimiter_index + 1:]
+                images.append(Image(data=image_data, type=cover_type,
+                                    desc=comment))
+            except KeyError:
+                pass
+
+        return images
+
+    def set_list(self, mutagen_file, values):
+        self.delete(mutagen_file)
+
+        for image in values:
+            image_type = image.type or ImageType.other
+            comment = image.desc or ''
+            image_data = comment + "\x00" + image.data
+            cover_tag = self.TAG_NAMES[image_type]
+            mutagen_file[cover_tag] = image_data
+
+    def delete(self, mutagen_file):
+        """Remove all images from the file.
+        """
+        for cover_tag in self.TAG_NAMES.values():
+            try:
+                del mutagen_file[cover_tag]
+            except KeyError:
+                pass
+
 
 # MediaField is a descriptor that represents a single logical field. It
 # aggregates several StorageStyles describing how to access the data for
@@ -1050,7 +1144,10 @@ class DateField(MediaField):
             return None
 
     def __set__(self, mediafile, date):
-        self._set_date_tuple(mediafile, date.year, date.month, date.day)
+        if date is None:
+            self._set_date_tuple(mediafile, None, None, None)
+        else:
+            self._set_date_tuple(mediafile, date.year, date.month, date.day)
 
     def __delete__(self, mediafile):
         super(DateField, self).__delete__(mediafile)
@@ -1062,11 +1159,11 @@ class DateField(MediaField):
         year, month, and day number. Each number is either an integer or
         None.
         """
-        # Get the underlying data and split on hyphens.
+        # Get the underlying data and split on hyphens and slashes.
         datestring = super(DateField, self).__get__(mediafile, None)
         if isinstance(datestring, basestring):
             datestring = re.sub(r'[Tt ].*$', '', unicode(datestring))
-            items = unicode(datestring).split('-')
+            items = re.split('[-/]', unicode(datestring))
         else:
             items = []
 
@@ -1096,13 +1193,15 @@ class DateField(MediaField):
         """
         if year is None:
             self.__delete__(mediafile)
-        date = [year]
+            return
+
+        date = [u'{0:04d}'.format(int(year))]
         if month:
-            date.append(month)
+            date.append(u'{0:02d}'.format(int(month)))
         if month and day:
-            date.append(day)
+            date.append(u'{0:02d}'.format(int(day)))
         date = map(unicode, date)
-        super(DateField, self).__set__(mediafile, '-'.join(date))
+        super(DateField, self).__set__(mediafile, u'-'.join(date))
 
         if hasattr(self, '_year_field'):
             self._year_field.__set__(mediafile, year)
@@ -1161,7 +1260,7 @@ class CoverArtField(MediaField):
         delattr(mediafile, 'images')
 
 
-class ImageListField(MediaField):
+class ImageListField(ListMediaField):
     """Descriptor to access the list of images embedded in tags.
 
     The getter returns a list of `Image` instances obtained from
@@ -1178,18 +1277,10 @@ class ImageListField(MediaField):
             ASFImageStorageStyle(),
             VorbisImageStorageStyle(),
             FlacImageStorageStyle(),
+            APEv2ImageStorageStyle(),
+            out_type=Image,
         )
 
-    def __get__(self, mediafile, _):
-        images = []
-        for style in self.styles(mediafile.mgfile):
-            images.extend(style.get_list(mediafile.mgfile))
-        return images
-
-    def __set__(self, mediafile, images):
-        for style in self.styles(mediafile.mgfile):
-            style.set_list(mediafile.mgfile, images)
-
 
 # MediaFile is a collection of fields.
 
@@ -1197,9 +1288,12 @@ class MediaFile(object):
     """Represents a multimedia file on disk and provides access to its
     metadata.
     """
-    def __init__(self, path):
+    def __init__(self, path, id3v23=False):
         """Constructs a new `MediaFile` reflecting the file at path. May
         throw `UnreadableFileError`.
+
+        By default, MP3 files are saved with ID3v2.4 tags. You can use
+        the older ID3v2.3 standard by specifying the `id3v23` option.
         """
         self.path = path
 
@@ -1214,12 +1308,13 @@ class MediaFile(object):
             mutagen.ogg.error,
             mutagen.asf.error,
             mutagen.apev2.error,
+            mutagen.aiff.error,
         )
         try:
             self.mgfile = mutagen.File(path)
         except unreadable_exc as exc:
             log.debug(u'header parsing failed: {0}'.format(unicode(exc)))
-            raise UnreadableFileError('Mutagen could not read file')
+            raise UnreadableFileError(path)
         except IOError as exc:
             if type(exc) == IOError:
                 # This is a base IOError, not a subclass from Mutagen or
@@ -1227,27 +1322,33 @@ class MediaFile(object):
                 raise
             else:
                 log.debug(traceback.format_exc())
-                raise MutagenError('Mutagen raised an exception')
+                raise MutagenError(path, exc)
         except Exception as exc:
             # Isolate bugs in Mutagen.
             log.debug(traceback.format_exc())
-            log.error('uncaught Mutagen exception in open: {0}'.format(exc))
-            raise MutagenError('Mutagen raised an exception')
+            log.error(u'uncaught Mutagen exception in open: {0}'.format(exc))
+            raise MutagenError(path, exc)
 
         if self.mgfile is None:
             # Mutagen couldn't guess the type
-            raise FileTypeError('file type unsupported by Mutagen')
+            raise FileTypeError(path)
         elif (type(self.mgfile).__name__ == 'M4A' or
               type(self.mgfile).__name__ == 'MP4'):
-            # This hack differentiates AAC and ALAC until we find a more
-            # deterministic approach. Mutagen only sets the sample rate
-            # for AAC files. See:
-            # https://github.com/sampsyo/beets/pull/295
-            if hasattr(self.mgfile.info, 'sample_rate') and \
-               self.mgfile.info.sample_rate > 0:
-                self.type = 'aac'
+            info = self.mgfile.info
+            if hasattr(info, 'codec'):
+                if info.codec and info.codec.startswith('alac'):
+                    self.type = 'alac'
+                else:
+                    self.type = 'aac'
             else:
-                self.type = 'alac'
+                # This hack differentiates AAC and ALAC on versions of
+                # Mutagen < 1.26. Once Mutagen > 1.26 is out and
+                # required by beets, we can remove this.
+                if hasattr(self.mgfile.info, 'bitrate') and \
+                   self.mgfile.info.bitrate > 0:
+                    self.type = 'aac'
+                else:
+                    self.type = 'alac'
         elif (type(self.mgfile).__name__ == 'ID3' or
               type(self.mgfile).__name__ == 'MP3'):
             self.type = 'mp3'
@@ -1265,22 +1366,24 @@ class MediaFile(object):
             self.type = 'mpc'
         elif type(self.mgfile).__name__ == 'ASF':
             self.type = 'asf'
+        elif type(self.mgfile).__name__ == 'AIFF':
+            self.type = 'aiff'
         else:
-            raise FileTypeError('file type %s unsupported by MediaFile' %
-                                type(self.mgfile).__name__)
+            raise FileTypeError(path, type(self.mgfile).__name__)
 
-        # add a set of tags if it's missing
+        # Add a set of tags if it's missing.
         if self.mgfile.tags is None:
             self.mgfile.add_tags()
 
-    def save(self, id3v23=False):
-        """Write the object's tags back to the file.
+        # Set the ID3v2.3 flag only for MP3s.
+        self.id3v23 = id3v23 and self.type == 'mp3'
 
-        By default, MP3 files are saved with ID3v2.4 tags. You can use
-        the older ID3v2.3 standard by specifying the `id3v23` option.
+    def save(self):
+        """Write the object's tags back to the file.
         """
+        # Possibly save the tags to ID3v2.3.
         kwargs = {}
-        if id3v23 and self.type == 'mp3':
+        if self.id3v23:
             id3 = self.mgfile
             if hasattr(id3, 'tags'):
                 # In case this is an MP3 object, not an ID3 object.
@@ -1296,8 +1399,8 @@ class MediaFile(object):
             raise
         except Exception as exc:
             log.debug(traceback.format_exc())
-            log.error('uncaught Mutagen exception in save: {0}'.format(exc))
-            raise MutagenError('Mutagen raised an exception')
+            log.error(u'uncaught Mutagen exception in save: {0}'.format(exc))
+            raise MutagenError(self.path, exc)
 
     def delete(self):
         """Remove the current metadata tag from the file.
@@ -1451,6 +1554,7 @@ class MediaFile(object):
         StorageStyle('DESCRIPTION'),
         StorageStyle('COMMENT'),
         ASFStorageStyle('WM/Comments'),
+        ASFStorageStyle('Description')
     )
     bpm = MediaField(
         MP3StorageStyle('TBPM'),
