@@ -20,9 +20,10 @@ from __future__ import print_function
 import logging
 import os
 import time
-import itertools
 import codecs
 import platform
+import re
+import shlex
 
 import beets
 from beets import ui
@@ -112,6 +113,30 @@ fields_cmd.func = fields_func
 default_commands.append(fields_cmd)
 
 
+# help: Print help text for commands
+
+class HelpCommand(ui.Subcommand):
+
+    def __init__(self):
+        super(HelpCommand, self).__init__(
+            'help', aliases=('?',),
+            help='give detailed help on a specific sub-command',
+        )
+
+    def func(self, lib, opts, args):
+        if args:
+            cmdname = args[0]
+            helpcommand = self.root_parser._subcommand_for_name(cmdname)
+            if not helpcommand:
+                raise ui.UserError("unknown command '{0}'".format(cmdname))
+            helpcommand.print_help()
+        else:
+            self.root_parser.print_help()
+
+
+default_commands.append(HelpCommand())
+
+
 # import: Autotagger and importer.
 
 # Importer utilities and support.
@@ -294,17 +319,9 @@ def show_change(cur_artist, cur_album, match):
                 color = 'lightgray'
             else:
                 color = 'red'
-            if (cur_track + new_track).count('-') == 1:
-                lhs_track, rhs_track = (ui.colorize(color, cur_track),
-                                        ui.colorize(color, new_track))
-            else:
-                color = 'red'
-                lhs_track, rhs_track = ui.color_diff_suffix(cur_track,
-                                                            new_track)
-            templ = (ui.colorize(color, u' (#') + u'{0}' +
-                     ui.colorize(color, u')'))
-            lhs += templ.format(lhs_track)
-            rhs += templ.format(rhs_track)
+            templ = ui.colorize(color, u' (#{0})')
+            lhs += templ.format(cur_track)
+            rhs += templ.format(new_track)
             lhs_width += len(cur_track) + 4
 
         # Length change.
@@ -313,12 +330,9 @@ def show_change(cur_artist, cur_album, match):
                 config['ui']['length_diff_thresh'].as_number():
             cur_length = ui.human_seconds_short(item.length)
             new_length = ui.human_seconds_short(track_info.length)
-            lhs_length, rhs_length = ui.color_diff_suffix(cur_length,
-                                                          new_length)
-            templ = (ui.colorize('red', u' (') + u'{0}' +
-                     ui.colorize('red', u')'))
-            lhs += templ.format(lhs_length)
-            rhs += templ.format(rhs_length)
+            templ = ui.colorize('red', u' ({0})')
+            lhs += templ.format(cur_length)
+            rhs += templ.format(new_length)
             lhs_width += len(cur_length) + 3
 
         # Penalties.
@@ -399,6 +413,37 @@ def show_item_change(item, match):
     print_(' '.join(info))
 
 
+def summarize_items(items, singleton):
+    """Produces a brief summary line describing a set of items. Used for
+    manually resolving duplicates during import.
+
+    `items` is a list of `Item` objects. `singleton` indicates whether
+    this is an album or single-item import (if the latter, them `items`
+    should only have one element).
+    """
+    summary_parts = []
+    if not singleton:
+        summary_parts.append("{0} items".format(len(items)))
+
+    format_counts = {}
+    for item in items:
+        format_counts[item.format] = format_counts.get(item.format, 0) + 1
+    if len(format_counts) == 1:
+        # A single format.
+        summary_parts.append(items[0].format)
+    else:
+        # Enumerate all the formats.
+        for format, count in format_counts.iteritems():
+            summary_parts.append('{0} {1}'.format(format, count))
+
+    average_bitrate = sum([item.bitrate for item in items]) / len(items)
+    total_duration = sum([item.length for item in items])
+    summary_parts.append('{0}kbps'.format(int(average_bitrate / 1000)))
+    summary_parts.append(ui.human_seconds_short(total_duration))
+
+    return ', '.join(summary_parts)
+
+
 def _summary_judment(rec):
     """Determines whether a decision should be made without even asking
     the user. This occurs in quiet mode and when an action is chosen for
@@ -716,18 +761,30 @@ class TerminalImportSession(importer.ImportSession):
                 assert isinstance(choice, autotag.TrackMatch)
                 return choice
 
-    def resolve_duplicate(self, task):
+    def resolve_duplicate(self, task, found_duplicates):
         """Decide what to do when a new album or item seems similar to one
         that's already in the library.
         """
-        log.warn("This %s is already in the library!" %
-                 ("album" if task.is_album else "item"))
+        log.warn(u"This {0} is already in the library!"
+                 .format("album" if task.is_album else "item"))
 
         if config['import']['quiet']:
             # In quiet mode, don't prompt -- just skip.
-            log.info('Skipping.')
+            log.info(u'Skipping.')
             sel = 's'
         else:
+            # Print some detail about the existing and new items so the
+            # user can make an informed decision.
+            for duplicate in found_duplicates:
+                print("Old: " + summarize_items(
+                    list(duplicate.items()) if task.is_album else [duplicate],
+                    not task.is_album,
+                ))
+            print("New: " + summarize_items(
+                task.imported_items(),
+                not task.is_album,
+            ))
+
             sel = ui.input_options(
                 ('Skip new', 'Keep both', 'Remove old')
             )
@@ -887,6 +944,10 @@ import_cmd.parser.add_option(
     '-g', '--group-albums', dest='group_albums', action='store_true',
     help='group tracks in a folder into separate albums'
 )
+import_cmd.parser.add_option(
+    '--pretend', dest='pretend', action='store_true',
+    help='just print the files to import'
+)
 import_cmd.func = import_func
 default_commands.append(import_cmd)
 
@@ -954,8 +1015,8 @@ def update_items(lib, query, album, move, pretend):
 
             # Did the item change since last checked?
             if item.current_mtime() <= item.mtime:
-                log.debug(u'skipping %s because mtime is up to date (%i)' %
-                          (displayable_path(item.path), item.mtime))
+                log.debug(u'skipping {0} because mtime is up to date ({1})'
+                          .format(displayable_path(item.path), item.mtime))
                 continue
 
             # Read new data.
@@ -1005,7 +1066,7 @@ def update_items(lib, query, album, move, pretend):
                 continue
             album = lib.get_album(album_id)
             if not album:  # Empty albums have already been removed.
-                log.debug('emptied album %i' % album_id)
+                log.debug(u'emptied album {0}'.format(album_id))
                 continue
             first_item = album.items().get()
 
@@ -1016,7 +1077,7 @@ def update_items(lib, query, album, move, pretend):
 
             # Move album art (and any inconsistent items).
             if move and lib.directory in ancestry(first_item.path):
-                log.debug('moving album %i' % album_id)
+                log.debug(u'moving album {0}'.format(album_id))
                 album.move()
 
 
@@ -1056,17 +1117,21 @@ def remove_items(lib, query, album, delete):
     # Get the matching items.
     items, albums = _do_query(lib, query, album)
 
-    # Show all the items.
-    for item in items:
-        ui.print_obj(item, lib)
-
-    # Confirm with user.
+    # Prepare confirmation with user.
     print_()
     if delete:
+        fmt = u'$path - $title'
         prompt = 'Really DELETE %i files (y/n)?' % len(items)
     else:
+        fmt = None
         prompt = 'Really remove %i items from the library (y/n)?' % \
                  len(items)
+
+    # Show all the items.
+    for item in items:
+        ui.print_obj(item, lib, fmt)
+
+    # Confirm with user.
     if not ui.input_yn(prompt, True):
         return
 
@@ -1106,6 +1171,7 @@ def show_stats(lib, query, exact):
     total_items = 0
     artists = set()
     albums = set()
+    album_artists = set()
 
     for item in items:
         if exact:
@@ -1115,18 +1181,29 @@ def show_stats(lib, query, exact):
         total_time += item.length
         total_items += 1
         artists.add(item.artist)
-        albums.add(item.album)
+        album_artists.add(item.albumartist)
+        if item.album_id:
+            albums.add(item.album_id)
 
     size_str = '' + ui.human_bytes(total_size)
     if exact:
         size_str += ' ({0} bytes)'.format(total_size)
 
     print_("""Tracks: {0}
-Total time: {1} ({2:.2f} seconds)
-Total size: {3}
-Artists: {4}
-Albums: {5}""".format(total_items, ui.human_seconds(total_time), total_time,
-                      size_str, len(artists), len(albums)))
+Total time: {1}{2}
+{3}: {4}
+Artists: {5}
+Albums: {6}
+Album artists: {7}""".format(
+        total_items,
+        ui.human_seconds(total_time),
+        ' ({0:.2f} seconds)'.format(total_time) if exact else '',
+        'Total size' if exact else 'Approximate total size',
+        size_str,
+        len(artists),
+        len(albums),
+        len(album_artists)),
+    )
 
 
 def stats_func(lib, opts, args):
@@ -1138,7 +1215,7 @@ stats_cmd = ui.Subcommand(
 )
 stats_cmd.parser.add_option(
     '-e', '--exact', action='store_true',
-    help='get exact file sizes'
+    help='exact size and time'
 )
 stats_cmd.func = stats_func
 default_commands.append(stats_cmd)
@@ -1167,15 +1244,16 @@ default_commands.append(version_cmd)
 
 def modify_items(lib, mods, dels, query, write, move, album, confirm):
     """Modifies matching items according to user-specified assignments and
-    deletions. `mods` is a list of "field=value" strings indicating
+    deletions.
+
+    `mods` is a dictionary of field and value pairse indicating
     assignments. `dels` is a list of fields to be deleted.
     """
     # Parse key=value specifications into a dictionary.
     model_cls = library.Album if album else library.Item
-    fsets = {}
-    for mod in mods:
-        key, value = mod.split('=', 1)
-        fsets[key] = model_cls._parse(key, value)
+
+    for key, value in mods.items():
+        mods[key] = model_cls._parse(key, value)
 
     # Get the items to modify.
     items, albums = _do_query(lib, query, album, False)
@@ -1183,13 +1261,16 @@ def modify_items(lib, mods, dels, query, write, move, album, confirm):
 
     # Apply changes *temporarily*, preview them, and collect modified
     # objects.
-    print_('Modifying %i %ss.' % (len(objs), 'album' if album else 'item'))
+    print_('Modifying {0} {1}s.'
+           .format(len(objs), 'album' if album else 'item'))
     changed = set()
     for obj in objs:
-        for field, value in fsets.iteritems():
-            obj[field] = value
+        obj.update(mods)
         for field in dels:
-            del obj[field]
+            try:
+                del obj[field]
+            except KeyError:
+                pass
         if ui.show_model_changes(obj):
             changed.add(obj)
 
@@ -1212,25 +1293,17 @@ def modify_items(lib, mods, dels, query, write, move, album, confirm):
         if not ui.input_yn('Really modify%s (Y/n)?' % extra):
             return
 
-    # Apply changes to database.
+    # Apply changes to database and files
     with lib.transaction():
         for obj in changed:
             if move:
                 cur_path = obj.path
                 if lib.directory in ancestry(cur_path):  # In library?
-                    log.debug('moving object %s' % cur_path)
+                    log.debug(u'moving object {0}'
+                              .format(displayable_path(cur_path)))
                     obj.move()
 
-            obj.store()
-
-    # Apply tags if requested.
-    if write:
-        if album:
-            changed_items = itertools.chain(*(a.items() for a in changed))
-        else:
-            changed_items = changed
-        for item in changed_items:
-            item.try_write()
+            obj.try_sync(write)
 
 
 def modify_parse_args(args):
@@ -1238,14 +1311,15 @@ def modify_parse_args(args):
     assignments (field=value), and deletions (field!).  Returns the result as
     a three-tuple in that order.
     """
-    mods = []
+    mods = {}
     dels = []
     query = []
     for arg in args:
         if arg.endswith('!') and '=' not in arg and ':' not in arg:
             dels.append(arg[:-1])  # Strip trailing !.
         elif '=' in arg and ':' not in arg.split('=', 1)[0]:
-            mods.append(arg)
+            key, val = arg.split('=', 1)
+            mods[key] = val
         else:
             query.append(arg)
     return query, mods, dels
@@ -1304,9 +1378,9 @@ def move_items(lib, dest, query, copy, album):
 
     action = 'Copying' if copy else 'Moving'
     entity = 'album' if album else 'item'
-    log.info('%s %i %ss.' % (action, len(objs), entity))
+    log.info(u'{0} {1} {2}s.'.format(action, len(objs), entity))
     for obj in objs:
-        log.debug('moving: %s' % obj.path)
+        log.debug(u'moving: {0}'.format(util.displayable_path(obj.path)))
 
         obj.move(copy, basedir=dest)
         obj.store()
@@ -1343,7 +1417,7 @@ default_commands.append(move_cmd)
 
 # write: Write tags into files.
 
-def write_items(lib, query, pretend):
+def write_items(lib, query, pretend, force):
     """Write tag information from the database to the respective files
     in the filesystem.
     """
@@ -1368,14 +1442,13 @@ def write_items(lib, query, pretend):
 
         # Check for and display changes.
         changed = ui.show_model_changes(item, clean_item,
-                                        library.Item._media_fields,
-                                        always=True)
-        if changed and not pretend:
-            item.try_write()
+                                        library.Item._media_fields, force)
+        if (changed or force) and not pretend:
+            item.try_sync()
 
 
 def write_func(lib, opts, args):
-    write_items(lib, decargs(args), opts.pretend)
+    write_items(lib, decargs(args), opts.pretend, opts.force)
 
 
 write_cmd = ui.Subcommand('write', help='write tag information to files')
@@ -1383,6 +1456,10 @@ write_cmd.parser.add_option(
     '-p', '--pretend', action='store_true',
     help="show all changes but do nothing"
 )
+write_cmd.parser.add_option(
+    '-f', '--force', action='store_true',
+    help="write tags even if the existing tags match the database"
+)
 write_cmd.func = write_func
 default_commands.append(write_cmd)
 
@@ -1413,32 +1490,43 @@ def config_func(lib, opts, args):
 
     # Open in editor.
     elif opts.edit:
-        path = config.user_config_path()
-
-        if 'EDITOR' in os.environ:
-            editor = os.environ['EDITOR']
-            args = [editor, editor, path]
-        elif platform.system() == 'Darwin':
-            args = ['open', 'open', '-n', path]
-        elif platform.system() == 'Windows':
-            # On windows we can execute arbitrary files. The os will
-            # take care of starting an appropriate application
-            args = [path, path]
-        else:
-            # Assume Unix
-            args = ['xdg-open', 'xdg-open', path]
-
-        try:
-            os.execlp(*args)
-        except OSError:
-            raise ui.UserError("Could not edit configuration. Please"
-                               "set the EDITOR environment variable.")
+        config_edit()
 
     # Dump configuration.
     else:
         print(config.dump(full=opts.defaults))
 
 
+def config_edit():
+    """Open a program to edit the user configuration.
+    """
+    path = config.user_config_path()
+
+    if 'EDITOR' in os.environ:
+        editor = os.environ['EDITOR']
+        try:
+            editor = shlex.split(editor)
+        except ValueError:  # Malformed shell tokens.
+            editor = [editor]
+        args = editor + [path]
+        args.insert(1, args[0])
+    elif platform.system() == 'Darwin':
+        args = ['open', 'open', '-n', path]
+    elif platform.system() == 'Windows':
+        # On windows we can execute arbitrary files. The os will
+        # take care of starting an appropriate application
+        args = [path, path]
+    else:
+        # Assume Unix
+        args = ['xdg-open', 'xdg-open', path]
+
+    try:
+        os.execlp(*args)
+    except OSError:
+        raise ui.UserError("Could not edit configuration. Please "
+                           "set the EDITOR environment variable.")
+
+
 config_cmd = ui.Subcommand('config',
                            help='show or edit the user configuration')
 config_cmd.parser.add_option(
@@ -1495,7 +1583,8 @@ def completion_script(commands):
         command_names.append(name)
 
         for alias in cmd.aliases:
-            aliases[alias] = name
+            if re.match(r'^\w+$', alias):
+                aliases[alias] = name
 
         options[name] = {'flags': [], 'opts': []}
         for opts in cmd.parser._get_all_options()[1:]:
@@ -1514,9 +1603,6 @@ def completion_script(commands):
         'opts': '-l --library -c --config -d --directory -h --help'.split(' ')
     }
 
-    # Help subcommand
-    command_names.append('help')
-
     # Add flags common to all commands
     options['_common'] = {
         'flags': ['-h', '--help']
