@@ -23,6 +23,7 @@ from string import Template
 import subprocess
 
 from beets import util, config, plugins, ui
+from beets.dbcore import types
 import pyechonest
 import pyechonest.song
 import pyechonest.track
@@ -38,7 +39,14 @@ DEVNULL = open(os.devnull, 'wb')
 ALLOWED_FORMATS = ('MP3', 'OGG', 'AAC')
 UPLOAD_MAX_SIZE = 50 * 1024 * 1024
 
-# The attributes we can import and where to store them in beets fields.
+# FIXME: use avconv?
+CONVERT_COMMAND = u'ffmpeg -i $source -y -acodec libvorbis -vn -aq 2 $dest'
+TRUNCATE_COMMAND = u'ffmpeg -t 300 -i $source'\
+                   u'-y -acodec libvorbis -vn -aq 2 $dest'
+
+# Maps attribute names from echonest to their field names in beets.
+# The attributes are retrieved from a songs `audio_summary`. See:
+# http://echonest.github.io/pyechonest/song.html#pyechonest.song.profile
 ATTRIBUTES = {
     'energy':       'energy',
     'liveness':     'liveness',
@@ -49,6 +57,16 @@ ATTRIBUTES = {
     'tempo':        'bpm',
 }
 
+# Types for the flexible fields added by `ATTRIBUTES`
+FIELD_TYPES = {
+    'energy':       types.FLOAT,
+    'liveness':     types.FLOAT,
+    'speechiness':  types.FLOAT,
+    'acousticness': types.FLOAT,
+    'danceability': types.FLOAT,
+    'valence':      types.FLOAT,
+}
+
 MUSICAL_SCALE = ['C', 'C#', 'D', 'D#', 'E' 'F',
                  'F#', 'G', 'G#', 'A', 'A#', 'B']
 
@@ -104,12 +122,14 @@ def similar(lib, src_item, threshold=0.15, fmt='${difference}: ${path}'):
 
 
 class EchonestMetadataPlugin(plugins.BeetsPlugin):
+
+    item_types = FIELD_TYPES
+
     def __init__(self):
         super(EchonestMetadataPlugin, self).__init__()
         self.config.add({
             'auto':    True,
             'apikey':  u'NY2KTZHQ0QDSHBAP6',
-            'codegen': None,
             'upload':  True,
             'convert': True,
             'truncate': True,
@@ -119,10 +139,6 @@ class EchonestMetadataPlugin(plugins.BeetsPlugin):
         pyechonest.config.ECHO_NEST_API_KEY = \
             config['echonest']['apikey'].get(unicode)
 
-        if config['echonest']['codegen']:
-            pyechonest.config.CODEGEN_BINARY_OVERRIDE = \
-                config['echonest']['codegen'].get(unicode)
-
         if self.config['auto']:
             self.import_stages = [self.imported]
 
@@ -137,6 +153,9 @@ class EchonestMetadataPlugin(plugins.BeetsPlugin):
             except pyechonest.util.EchoNestAPIError as e:
                 if e.code == 3:
                     # reached access limit per minute
+                    log.debug(u'echonest: rate-limited on try {0}; '
+                              u'waiting {1} seconds'
+                              .format(i + 1, RETRY_INTERVAL))
                     time.sleep(RETRY_INTERVAL)
                 elif e.code == 5:
                     # specified identifier does not exist
@@ -158,7 +177,7 @@ class EchonestMetadataPlugin(plugins.BeetsPlugin):
         else:
             # If we exited the loop without breaking, then we used up all
             # our allotted retries.
-            raise ui.UserError(u'echonest request failed repeatedly')
+            log.error(u'echonest request failed repeatedly')
             return None
         return result
 
@@ -233,69 +252,53 @@ class EchonestMetadataPlugin(plugins.BeetsPlugin):
                                        'audio_summary'])
         return self._flatten_song(self._pick_song(songs, item))
 
-    # "Identify" (fingerprinting) lookup.
-
-    def fingerprint(self, item):
-        """Get the fingerprint for this item from the EchoNest.  If we
-        already have a fingerprint, return it and don't calculate it
-        again.
-        """
-        if FINGERPRINT_KEY in item:
-            return item[FINGERPRINT_KEY]
-
-        try:
-            res = self._echofun(pyechonest.util.codegen,
-                                filename=item.path.decode('utf-8'))
-        except Exception as e:
-            # Frustratingly, the pyechonest library raises a plain Exception
-            # when the command is not found.
-            log.debug(u'echonest: codegen failed: {0}'.format(e))
-            return
-
-        if not res or 'code' not in res[0] or not res[0]['code']:
-            log.debug(u'echonest: no fingerprint returned')
-            return
-        code = res[0]['code']
-
-        log.debug(u'echonest: calculated fingerprint')
-        item[FINGERPRINT_KEY] = code
-        return code
-
-    def identify(self, item):
-        """Try to identify the song at the EchoNest.
-        """
-        code = self.fingerprint(item)
-        if not code:
-            return
-
-        songs = self._echofun(pyechonest.song.identify, code=code)
-        if not songs:
-            log.debug(u'echonest: no songs found for fingerprint')
-            return
-
-        return self._flatten_song(max(songs, key=lambda s: s.score))
-
     # "Analyze" (upload the audio itself) method.
 
-    def convert(self, item):
+    def prepare_upload(self, item):
+        """Truncate and convert an item's audio file so it can be
+        uploaded to echonest.
+
+        Return a ``(source, tmp)`` tuple where `source` is the path to
+        the file to be uploaded and `tmp` is a temporary file to be
+        deleted after the upload or `None`.
+
+        If conversion or truncation fails, return `None`.
+        """
+        source = item.path
+        tmp = None
+        if item.format not in ALLOWED_FORMATS:
+            if config['echonest']['convert']:
+                tmp = source = self.convert(source)
+            if not tmp:
+                return
+
+        if os.stat(source).st_size > UPLOAD_MAX_SIZE:
+            if config['echonest']['truncate']:
+                source = self.truncate(source)
+                if tmp is not None:
+                    util.remove(tmp)
+                tmp = source
+            else:
+                return
+
+        if source:
+            return source, tmp
+
+    def convert(self, source):
         """Converts an item in an unsupported media format to ogg.  Config
         pending.
         This is stolen from Jakob Schnitzers convert plugin.
         """
         fd, dest = tempfile.mkstemp(u'.ogg')
         os.close(fd)
-        source = item.path
 
         log.info(u'echonest: encoding {0} to {1}'.format(
             util.displayable_path(source),
             util.displayable_path(dest),
         ))
 
-        # Build up the FFmpeg command line.
-        # FIXME: use avconv?
-        command = u'ffmpeg -i $source -y -acodec libvorbis -vn -aq 2 $dest'
         opts = []
-        for arg in command.split():
+        for arg in CONVERT_COMMAND.split():
             arg = arg.encode('utf-8')
             opts.append(Template(arg).substitute(source=source, dest=dest))
 
@@ -312,20 +315,18 @@ class EchonestMetadataPlugin(plugins.BeetsPlugin):
         )
         return dest
 
-    def truncate(self, item):
+    def truncate(self, source):
         """Truncates an item to a size less than UPLOAD_MAX_SIZE."""
         fd, dest = tempfile.mkstemp(u'.ogg')
         os.close(fd)
-        source = item.path
 
         log.info(u'echonest: truncating {0} to {1}'.format(
             util.displayable_path(source),
             util.displayable_path(dest),
         ))
 
-        command = u'ffmpeg -t 300 -i $source -y -acodec copy $dest'
         opts = []
-        for arg in command.split():
+        for arg in TRUNCATE_COMMAND.split():
             arg = arg.encode('utf-8')
             opts.append(Template(arg).substitute(source=source, dest=dest))
 
@@ -346,31 +347,18 @@ class EchonestMetadataPlugin(plugins.BeetsPlugin):
         """Upload the item to the EchoNest for analysis. May require to
         convert the item to a supported media format.
         """
-        # Get the file to upload (either by using the file directly or by
-        # transcoding it first).
-        source = item.path
-        if item.format not in ALLOWED_FORMATS:
-            if config['echonest']['convert']:
-                source = self.convert(item)
-                if not source:
-                    log.debug(u'echonest: failed to convert file')
-                    return
-            else:
-                return
+        prepared = self.prepare_upload(item)
+        if not prepared:
+            log.debug(u'echonest: could not prepare file for upload')
+            return
 
-        if os.stat(item.path).st_size > UPLOAD_MAX_SIZE:
-            if config['echonest']['truncate']:
-                source = self.truncate(item)
-                if not source:
-                    log.debug(u'echonest: failed to truncate file')
-                    return
-            else:
-                return
-
-        # Upload the audio file.
+        source, tmp = prepared
         log.info(u'echonest: uploading file, please be patient')
         track = self._echofun(pyechonest.track.track_from_filename,
                               filename=source)
+        if tmp is not None:
+            util.remove(tmp)
+
         if not track:
             log.debug(u'echonest: failed to upload file')
             return
@@ -416,8 +404,6 @@ class EchonestMetadataPlugin(plugins.BeetsPlugin):
         # There are four different ways to get a song. Each method is a
         # callable that takes the Item as an argument.
         methods = [self.profile, self.search]
-        if config['echonest']['codegen']:
-            methods.append(self.identify)
         if config['echonest']['upload']:
             methods.append(self.analyze)
 
@@ -430,7 +416,7 @@ class EchonestMetadataPlugin(plugins.BeetsPlugin):
                         method.__name__,
                         item.artist,
                         item.title,
-                        song['duration'],
+                        song.get('duration'),
                     )
                 )
                 return song
