@@ -25,6 +25,7 @@ from beets import config
 from beets import plugins
 from beets import library
 from beets.util import displayable_path
+from beets.dbcore import types
 
 log = logging.getLogger('beets')
 
@@ -245,6 +246,10 @@ class MPDStats(object):
 
     def on_stop(self, status):
         log.info(u'mpdstats: stop')
+
+        if self.now_playing:
+            self.handle_song_change(self.now_playing)
+
         self.now_playing = None
 
     def on_pause(self, status):
@@ -304,6 +309,14 @@ class MPDStats(object):
 
 
 class MPDStatsPlugin(plugins.BeetsPlugin):
+
+    item_types = {
+        'play_count':  types.INTEGER,
+        'skip_count':  types.INTEGER,
+        'last_played': library.DateType(),
+        'rating':      types.FLOAT,
+    }
+
     def __init__(self):
         super(MPDStatsPlugin, self).__init__()
         self.config.add({
