@@ -25,7 +25,7 @@ from beets.plugins import BeetsPlugin
 from beets.util import syspath, command_output, displayable_path
 from beets import config
 
-log = logging.getLogger('beets')
+log = logging.getLogger('beets.replaygain')
 
 
 # Utilities.
@@ -135,7 +135,7 @@ class CommandBackend(Backend):
 
         supported_items = filter(self.format_supported, album.items())
         if len(supported_items) != len(album.items()):
-            log.debug('replaygain: tracks are of unsupported format')
+            log.debug(u'replaygain: tracks are of unsupported format')
             return AlbumGain(None, [])
 
         output = self.compute_gain(supported_items, True)
@@ -198,6 +198,9 @@ class CommandBackend(Backend):
         out = []
         for line in text.split('\n')[1:num_lines + 1]:
             parts = line.split('\t')
+            if len(parts) != 6 or parts[0] == 'File':
+                log.debug(u'replaygain: bad tool output: {0}'.format(text))
+                raise ReplayGainError('mp3gain failed')
             d = {
                 'file': parts[0],
                 'mp3gain': int(parts[1]),
@@ -270,23 +273,32 @@ class GStreamerBackend(object):
 
         try:
             import gi
-            gi.require_version('Gst', '1.0')
+        except ImportError:
+            raise FatalReplayGainError(
+                "Failed to load GStreamer: python-gi not found"
+            )
 
-            from gi.repository import GObject, Gst, GLib
-            # Calling GObject.threads_init() is not needed for
-            # PyGObject 3.10.2+
-            with warnings.catch_warnings():
-                warnings.simplefilter("ignore")
-                GObject.threads_init()
-            Gst.init([sys.argv[0]])
-        except:
-            raise FatalReplayGainError("GStreamer failed to initialize")
+        try:
+            gi.require_version('Gst', '1.0')
+        except ValueError as e:
+            raise FatalReplayGainError(
+                "Failed to load GStreamer 1.0: {0}".format(e)
+            )
+
+        from gi.repository import GObject, Gst, GLib
+        # Calling GObject.threads_init() is not needed for
+        # PyGObject 3.10.2+
+        with warnings.catch_warnings():
+            warnings.simplefilter("ignore")
+            GObject.threads_init()
+        Gst.init([sys.argv[0]])
 
         self.GObject = GObject
         self.GLib = GLib
         self.Gst = Gst
 
     def compute(self, files, album):
+        self._error = None
         self._files = list(files)
 
         if len(self._files) == 0:
@@ -299,6 +311,8 @@ class GStreamerBackend(object):
 
         if self._set_first_file():
             self._main_loop.run()
+            if self._error is not None:
+                raise self._error
 
     def compute_track_gain(self, items):
         self.compute(items, False)
@@ -342,9 +356,12 @@ class GStreamerBackend(object):
         self._pipe.set_state(self.Gst.State.NULL)
         self._main_loop.quit()
         err, debug = message.parse_error()
+        f = self._src.get_property("location")
         # A GStreamer error, either an unsupported format or a bug.
-        raise ReplayGainError("Error %s - %s on file %s" %
-                              (err, debug, self._src.get_property("location")))
+        self._error = \
+            ReplayGainError(u"Error {0} - {1} on file {2}".format(err,
+                                                                  debug,
+                                                                  f))
 
     def _on_tag(self, bus, message):
         tags = message.parse_tag()
@@ -446,6 +463,154 @@ class GStreamerBackend(object):
         assert(peer is None)
 
 
+class AudioToolsBackend(Backend):
+    """ReplayGain backend that uses `Python Audio Tools
+    <http://audiotools.sourceforge.net/>`_ and its capabilities to read more
+    file formats and compute ReplayGain values using it replaygain module.
+    """
+    def __init__(self, config):
+        self._import_audiotools()
+
+    def _import_audiotools(self):
+        """Check whether it's possible to import the necessary modules.
+        There is no check on the file formats at runtime.
+
+        :raises :exc:`ReplayGainError`: if the modules cannot be imported
+        """
+        try:
+            import audiotools
+            import audiotools.replaygain
+        except ImportError:
+            raise FatalReplayGainError(
+                "Failed to load audiotools: audiotools not found"
+            )
+        self._mod_audiotools = audiotools
+        self._mod_replaygain = audiotools.replaygain
+
+    def open_audio_file(self, item):
+        """Open the file to read the PCM stream from the using
+        ``item.path``.
+
+        :return: the audiofile instance
+        :rtype: :class:`audiotools.AudioFile`
+        :raises :exc:`ReplayGainError`: if the file is not found or the
+        file format is not supported
+        """
+        try:
+            audiofile = self._mod_audiotools.open(item.path)
+        except IOError:
+            raise ReplayGainError(
+                "File {} was not found".format(item.path)
+            )
+        except self._mod_audiotools.UnsupportedFile:
+            raise ReplayGainError(
+                "Unsupported file type {}".format(item.format)
+            )
+
+        return audiofile
+
+    def init_replaygain(self, audiofile, item):
+        """Return an initialized :class:`audiotools.replaygain.ReplayGain`
+        instance, which requires the sample rate of the song(s) on which
+        the ReplayGain values will be computed. The item is passed in case
+        the sample rate is invalid to log the stored item sample rate.
+
+        :return: initialized replagain object
+        :rtype: :class:`audiotools.replaygain.ReplayGain`
+        :raises: :exc:`ReplayGainError` if the sample rate is invalid
+        """
+        try:
+            rg = self._mod_replaygain.ReplayGain(audiofile.sample_rate())
+        except ValueError:
+            raise ReplayGainError(
+                "Unsupported sample rate {}".format(item.samplerate)
+            )
+            return
+        return rg
+
+    def compute_track_gain(self, items):
+        """Compute ReplayGain values for the requested items.
+
+        :return list: list of :class:`Gain` objects
+        """
+        return [self._compute_track_gain(item) for item in items]
+
+    def _compute_track_gain(self, item):
+        """Compute ReplayGain value for the requested item.
+
+        :rtype: :class:`Gain`
+        """
+        audiofile = self.open_audio_file(item)
+        rg = self.init_replaygain(audiofile, item)
+        # Each call to title_gain on a replaygain object return peak and gain
+        # of the track.
+        # Note that the method needs an audiotools.PCMReader instance that can
+        # be obtained from an audiofile instance.
+        rg_track_gain, rg_track_peak = rg.title_gain(audiofile.to_pcm())
+
+        log.debug(
+            u'ReplayGain for track {0} - {1}: {2:.2f}, {3:.2f}'.format(
+                item.artist,
+                item.title,
+                rg_track_gain,
+                rg_track_peak
+            )
+        )
+        return Gain(gain=rg_track_gain, peak=rg_track_peak)
+
+    def compute_album_gain(self, album):
+        """Compute ReplayGain values for the requested album and its items.
+
+        :rtype: :class:`AlbumGain`
+        """
+        log.debug(
+            u'Analysing album {0} - {1}'.format(
+                album.albumartist,
+                album.album
+            )
+        )
+
+        # The first item is taken and opened to get the sample rate to
+        # initialize the replaygain object. The object is used for all the
+        # tracks in the album to get the album values.
+        item = list(album.items())[0]
+        audiofile = self.open_audio_file(item)
+        rg = self.init_replaygain(audiofile, item)
+
+        track_gains = []
+        for item in album.items():
+            audiofile = self.open_audio_file(item)
+            rg_track_gain, rg_track_peak = rg.title_gain(audiofile.to_pcm())
+            track_gains.append(
+                Gain(gain=rg_track_gain, peak=rg_track_peak)
+            )
+            log.debug(
+                u'ReplayGain for track {0} - {1}: {2:.2f}, {3:.2f}'.format(
+                    item.artist,
+                    item.title,
+                    rg_track_gain,
+                    rg_track_peak
+                )
+            )
+
+        # After getting the values for all tracks, it's possible to get the
+        # album values.
+        rg_album_gain, rg_album_peak = rg.album_gain()
+        log.debug(
+            u'ReplayGain for Album {0} - {1}: {2:.2f}, {3:.2f}'.format(
+                album.albumartist,
+                album.album,
+                rg_album_gain,
+                rg_album_peak
+            )
+        )
+
+        return AlbumGain(
+            Gain(gain=rg_album_gain, peak=rg_album_peak),
+            track_gains=track_gains
+        )
+
+
 # Main plugin logic.
 
 class ReplayGainPlugin(BeetsPlugin):
@@ -455,6 +620,7 @@ class ReplayGainPlugin(BeetsPlugin):
     backends = {
         "command":   CommandBackend,
         "gstreamer": GStreamerBackend,
+        "audiotools": AudioToolsBackend
     }
 
     def __init__(self):
@@ -487,7 +653,7 @@ class ReplayGainPlugin(BeetsPlugin):
             )
         except (ReplayGainError, FatalReplayGainError) as e:
             raise ui.UserError(
-                'An error occurred in backend initialization: {0}'.format(e)
+                'replaygain initialization failed: {0}'.format(e)
             )
 
     def track_requires_gain(self, item):
@@ -555,7 +721,7 @@ class ReplayGainPlugin(BeetsPlugin):
                 if write:
                     item.try_write()
         except ReplayGainError as e:
-            log.warn(u"ReplayGain error: {0}".format(e))
+            log.info(u"ReplayGain error: {0}".format(e))
         except FatalReplayGainError as e:
             raise ui.UserError(
                 u"Fatal replay gain error: {0}".format(e)
@@ -569,12 +735,12 @@ class ReplayGainPlugin(BeetsPlugin):
         in the item, nothing is done.
         """
         if not self.track_requires_gain(item):
-            log.info(u'Skipping track {0} - {1}'.format(item.artist,
-                                                        item.title))
+            log.info(u'Skipping track {0} - {1}'
+                     .format(item.artist, item.title))
             return
 
-        log.info(u'analyzing {0} - {1}'.format(item.artist,
-                                               item.title))
+        log.info(u'analyzing {0} - {1}'
+                 .format(item.artist, item.title))
 
         try:
             track_gains = self.backend_instance.compute_track_gain([item])
@@ -589,7 +755,7 @@ class ReplayGainPlugin(BeetsPlugin):
             if write:
                 item.try_write()
         except ReplayGainError as e:
-            log.warn(u"ReplayGain error: {0}".format(e))
+            log.info(u"ReplayGain error: {0}".format(e))
         except FatalReplayGainError as e:
             raise ui.UserError(
                 u"Fatal replay gain error: {0}".format(e)
@@ -601,6 +767,8 @@ class ReplayGainPlugin(BeetsPlugin):
         if not self.automatic:
             return
 
+        log.setLevel(logging.WARN)
+
         if task.is_album:
             self.handle_album(task.album, False)
         else:
@@ -610,6 +778,8 @@ class ReplayGainPlugin(BeetsPlugin):
         """Return the "replaygain" ui subcommand.
         """
         def func(lib, opts, args):
+            log.setLevel(logging.INFO)
+
             write = config['import']['write'].get(bool)
 
             if opts.album:
